@@ -1,44 +1,111 @@
 """Rule for checking content of jinja template strings."""
+
 from __future__ import annotations
 
 import logging
+import os
 import re
 import sys
-from collections import namedtuple
-from typing import TYPE_CHECKING, Any
+from collections.abc import Mapping
+from dataclasses import dataclass
+from pathlib import Path
+from typing import TYPE_CHECKING, NamedTuple
 
 import black
 import jinja2
 from ansible.errors import AnsibleError, AnsibleFilterError, AnsibleParserError
-from ansible.parsing.yaml.objects import AnsibleUnicode
 from jinja2.exceptions import TemplateSyntaxError
 
-from ansiblelint.constants import LINE_NUMBER_KEY
+from ansiblelint.errors import RuleMatchTransformMeta
 from ansiblelint.file_utils import Lintable
-from ansiblelint.rules import AnsibleLintRule
+from ansiblelint.rules import AnsibleLintRule, TransformMixin
+from ansiblelint.runner import get_matches
 from ansiblelint.skip_utils import get_rule_skips_from_line
-from ansiblelint.utils import parse_yaml_from_file, template
+from ansiblelint.text import has_jinja
+from ansiblelint.types import AnsibleTemplateSyntaxError
+from ansiblelint.utils import (  # type: ignore[attr-defined]
+    Templar,
+    parse_yaml_from_file,
+    template,
+)
 from ansiblelint.yaml_utils import deannotate, nested_items_path
 
 if TYPE_CHECKING:
+    from ruamel.yaml.comments import CommentedMap, CommentedSeq
+
+    from ansiblelint.config import Options
     from ansiblelint.errors import MatchError
+    from ansiblelint.utils import Task
 
 
 _logger = logging.getLogger(__package__)
 KEYWORDS_WITH_IMPLICIT_TEMPLATE = ("changed_when", "failed_when", "until", "when")
 
-Token = namedtuple("Token", "lineno token_type value")
+
+class Token(NamedTuple):
+    """Token."""
+
+    lineno: int
+    token_type: str
+    value: str
 
 
-class JinjaRule(AnsibleLintRule):
+ignored_re = re.compile(
+    "|".join(  # noqa: FLY002
+        [
+            r"^Object of type method is not JSON serializable",
+            r"^Unexpected templating type error occurred on",
+            r"^obj must be a list of dicts or a nested dict$",
+            r"^the template file (.*) could not be found for the lookup$",
+            r"could not locate file in lookup",
+            r"unable to locate collection",
+            r"^Error in (.*)is undefined$",
+            r"^Mandatory variable (.*) not defined.$",
+            r"is undefined",
+            r"Unrecognized type <<class 'ansible.template.AnsibleUndefined'>> for (.*) filter <value>$",
+            # https://github.com/ansible/ansible-lint/issues/3155
+            r"^The '(.*)' test expects a dictionary$",
+            # https://github.com/ansible/ansible-lint/issues/4338
+            r"An unhandled exception occurred while templating (.*). Error was a <class 'ansible.errors.AnsibleFilterError'>, original message: The (.*) test expects a dictionary$",
+        ],
+    ),
+    flags=re.MULTILINE | re.DOTALL,
+)
+
+
+@dataclass(frozen=True)
+class JinjaRuleTMetaSpacing(RuleMatchTransformMeta):
+    """JinjaRule transform metadata.
+
+    :param key: Key or index within the task
+    :param value: Value of the key
+    :param path: Path to the key
+    :param fixed: Value with spacing fixed
+    """
+
+    key: str | int
+    value: str | int
+    path: tuple[str | int, ...]
+    fixed: str
+
+    def __str__(self) -> str:
+        """Return string representation."""
+        return f"{self.key}={self.value} at {self.path} fixed to {self.fixed}"
+
+
+class JinjaRule(AnsibleLintRule, TransformMixin):
     """Rule that looks inside jinja2 templates."""
 
     id = "jinja"
     severity = "LOW"
     tags = ["formatting"]
-    version_added = "v6.5.0"
+    version_changed = "6.5.0"
     _ansible_error_re = re.compile(
-        r"^(?P<error>.*): (?P<detail>.*)\. String: (?P<string>.*)$", flags=re.MULTILINE
+        (
+            r"^(?P<error>.*): (?P<detail>.*)\. String: (?P<string>.*)$"
+            r"|An unhandled exception occurred while templating '.*'\. Error was a .*, original message: (?P<nested_error>.*)"
+        ),
+        flags=re.MULTILINE,
     )
 
     env = jinja2.Environment(trim_blocks=False)
@@ -46,123 +113,173 @@ class JinjaRule(AnsibleLintRule):
         "invalid": "Syntax error in jinja2 template: {value}",
         "spacing": "Jinja2 spacing could be improved: {value} -> {reformatted}",
     }
+    _ids = {
+        "jinja[invalid]": "Invalid jinja2 syntax",
+        "jinja[spacing]": "Jinja2 spacing could be improved",
+    }
 
     def _msg(self, tag: str, value: str, reformatted: str) -> str:
         """Generate error message."""
         return self._tag2msg[tag].format(value=value, reformatted=reformatted)
 
-    # pylint: disable=too-many-branches
-    def matchtask(  # noqa: C901
-        self, task: dict[str, Any], file: Lintable | None = None
-    ) -> bool | str | MatchError:
-        for key, v, _ in nested_items_path(task):
-            if isinstance(v, str):
-
-                try:
-                    template(
-                        basedir=file.dir if file else ".",
-                        value=v,
-                        variables=deannotate(task.get("vars", {})),
-                        fail_on_error=True,  # we later decide which ones to ignore or not
-                    )
-                # ValueError RepresenterError
-                except AnsibleError as exc:
-                    bypass = False
-                    orig_exc = exc.orig_exc if getattr(exc, "orig_exc", None) else exc
-                    match = self._ansible_error_re.match(
-                        getattr(orig_exc, "message", str(orig_exc))
-                    )
-                    if (
-                        isinstance(exc, AnsibleFilterError)
-                        or "unable to locate collection" in orig_exc.message
-                    ):
-                        bypass = True
-                    elif re.match(
-                        r"^the template file (.*) could not be found for the lookup$",
-                        orig_exc.message,
-                    ) or re.match(r"could not locate file in lookup", orig_exc.message):
-                        bypass = True
-                    elif isinstance(orig_exc, AnsibleParserError):
-                        # "An unhandled exception occurred while running the lookup plugin '...'. Error was a <class 'ansible.errors.AnsibleParserError'>, original message: Invalid filename: 'None'. Invalid filename: 'None'"
-
-                        # An unhandled exception occurred while running the lookup plugin 'template'. Error was a <class 'ansible.errors.AnsibleError'>, original message: the template file ... could not be found for the lookup. the template file ... could not be found for the lookup
-
-                        # ansible@devel (2.14) new behavior:
-                        # AnsibleError(TemplateSyntaxError): template error while templating string: Could not load "ipwrap": 'Invalid plugin FQCN (ansible.netcommon.ipwrap): unable to locate collection ansible.netcommon'. String: Foo {{ buildset_registry.host | ipwrap }}. Could not load "ipwrap": 'Invalid plugin FQCN (ansible.netcommon.ipwrap): unable to locate collection ansible.netcommon'
-                        bypass = True
-                    elif (
-                        isinstance(orig_exc, (AnsibleError, TemplateSyntaxError))
-                        and match
-                    ):
-                        error = match.group("error")
-                        detail = match.group("detail")
-                        # string = match.group("string")
-                        if error.startswith("template error while templating string"):
-                            bypass = False
-                        elif detail.startswith("unable to locate collection"):
-                            _logger.debug("Ignored AnsibleError: %s", exc)
-                            bypass = True
-                        else:
-                            bypass = False
-                    elif re.match(r"^lookup plugin (.*) not found$", exc.message):
-                        # lookup plugin 'template' not found
-                        bypass = True
-
-                    # AnsibleFilterError: 'obj must be a list of dicts or a nested dict'
-                    # AnsibleError: template error while templating string: expected token ':', got '}'. String: {{ {{ '1' }} }}
-                    # AnsibleError: template error while templating string: unable to locate collection ansible.netcommon. String: Foo {{ buildset_registry.host | ipwrap }}
-                    if not bypass:
-                        return self.create_matcherror(
-                            message=str(exc),
-                            linenumber=task[LINE_NUMBER_KEY],
-                            filename=file,
-                            tag=f"{self.id}[invalid]",
+    # pylint: disable=too-many-locals
+    def matchtask(
+        self,
+        task: Task,
+        file: Lintable | None = None,
+    ) -> list[MatchError]:
+        result = []
+        try:
+            for key, v, path in nested_items_path(
+                task,
+                ignored_keys=("block", "ansible.builtin.block", "ansible.legacy.block"),
+            ):
+                if isinstance(v, str):
+                    try:
+                        template(
+                            basedir=file.path.parent if file else Path(),
+                            value=v,
+                            variables=deannotate(task.get("vars", {})),
+                            fail_on_error=True,  # we later decide which ones to ignore or not
                         )
+                    except AnsibleFilterError:
+                        bypass = True
+                    # ValueError RepresenterError
+                    except AnsibleError as exc:
+                        bypass = False
+                        orig_exc = (
+                            exc.orig_exc if getattr(exc, "orig_exc", None) else exc
+                        )
+                        orig_exc_message = getattr(orig_exc, "message", str(orig_exc))
+                        match = self._ansible_error_re.match(
+                            getattr(orig_exc, "message", str(orig_exc)),
+                        )
+                        if ignored_re.search(orig_exc_message) or isinstance(
+                            orig_exc,
+                            AnsibleParserError | TypeError,
+                        ):
+                            # An unhandled exception occurred while running the lookup plugin 'template'. Error was a <class 'ansible.errors.AnsibleError'>, original message: the template file ... could not be found for the lookup. the template file ... could not be found for the lookup
 
-                reformatted, details, tag = self.check_whitespace(
-                    v, key=key, lintable=file
-                )
-                if reformatted != v:
-                    return self.create_matcherror(
-                        message=self._msg(tag=tag, value=v, reformatted=reformatted),
-                        linenumber=task[LINE_NUMBER_KEY],
-                        details=details,
-                        filename=file,
-                        tag=f"{self.id}[{tag}]",
+                            # ansible@devel (2.14) new behavior:
+                            # AnsibleError(TemplateSyntaxError): template error while templating string: Could not load "ipwrap": 'Invalid plugin FQCN (ansible.netcommon.ipwrap): unable to locate collection ansible.netcommon'. String: Foo {{ buildset_registry.host | ipwrap }}. Could not load "ipwrap": 'Invalid plugin FQCN (ansible.netcommon.ipwrap): unable to locate collection ansible.netcommon'
+                            bypass = True
+                        elif (
+                            isinstance(orig_exc, AnsibleError | TemplateSyntaxError)
+                            and match
+                        ):
+                            error = match.group("error")
+                            detail = match.group("detail")
+                            nested_error = match.group("nested_error")
+                            if error and error.startswith(
+                                "template error while templating string",
+                            ):
+                                bypass = False
+                            elif detail and detail.startswith(
+                                "unable to locate collection",
+                            ):
+                                _logger.debug("Ignored AnsibleError: %s", exc)
+                                bypass = True
+                            elif nested_error and nested_error.startswith(
+                                "Unexpected templating type error occurred on",
+                            ):
+                                bypass = True
+                            else:
+                                bypass = False
+                        elif re.match(r"^lookup plugin (.*) not found$", exc.message):
+                            # lookup plugin 'template' not found
+                            bypass = True
+                        elif isinstance(
+                            orig_exc, AnsibleTemplateSyntaxError
+                        ) and re.match(
+                            r"^Syntax error in template: No filter named '.*'.",
+                            exc.message,
+                        ):
+                            bypass = True
+
+                        # AnsibleError: template error while templating string: expected token ':', got '}'. String: {{ {{ '1' }} }}
+                        # AnsibleError: template error while templating string: unable to locate collection ansible.netcommon. String: Foo {{ buildset_registry.host | ipwrap }}
+                        if not bypass:
+                            lineno = task.get_error_line([*path, key])
+                            result.append(
+                                self.create_matcherror(
+                                    message=str(exc),
+                                    lineno=lineno,
+                                    data=v,
+                                    filename=file,
+                                    tag=f"{self.id}[invalid]",
+                                ),
+                            )
+                            continue
+                    reformatted, details, tag = self.check_whitespace(
+                        v,
+                        key=key,
+                        lintable=file,
                     )
-        return False
+                    if reformatted != v:
+                        lineno = task.get_error_line([*path, key])
+                        result.append(
+                            self.create_matcherror(
+                                message=self._msg(
+                                    tag=tag,
+                                    value=v,
+                                    reformatted=reformatted,
+                                ),
+                                lineno=lineno,
+                                data=v,
+                                details=details,
+                                filename=file,
+                                tag=f"{self.id}[{tag}]",
+                                transform_meta=JinjaRuleTMetaSpacing(
+                                    key=key,
+                                    value=v,
+                                    path=tuple(path),
+                                    fixed=reformatted,
+                                ),
+                            ),
+                        )
+        except Exception as exc:
+            _logger.info("Exception in JinjaRule.matchtask: %s", exc)
+            raise
+        return result
 
     def matchyaml(self, file: Lintable) -> list[MatchError]:
         """Return matches for variables defined in vars files."""
-        data: dict[str, Any] = {}
         raw_results: list[MatchError] = []
         results: list[MatchError] = []
 
         if str(file.kind) == "vars":
             data = parse_yaml_from_file(str(file.path))
-            # pylint: disable=unused-variable
-            for key, v, path in nested_items_path(data):
-                if isinstance(v, AnsibleUnicode):
+            if not isinstance(data, Mapping):
+                return results
+            for key, v, _path in nested_items_path(data):
+                if isinstance(v, str):
                     reformatted, details, tag = self.check_whitespace(
-                        v, key=key, lintable=file
+                        v,
+                        key=key,
+                        lintable=file,
                     )
                     if reformatted != v:
                         results.append(
                             self.create_matcherror(
                                 message=self._msg(
-                                    tag=tag, value=v, reformatted=reformatted
+                                    tag=tag,
+                                    value=v,
+                                    reformatted=reformatted,
                                 ),
-                                linenumber=v.ansible_pos[1],
+                                data=v,
                                 details=details,
                                 filename=file,
                                 tag=f"{self.id}[{tag}]",
-                            )
+                            ),
                         )
             if raw_results:
                 lines = file.content.splitlines()
                 for match in raw_results:
-                    # linenumber starts with 1, not zero
-                    skip_list = get_rule_skips_from_line(lines[match.linenumber - 1])
+                    # lineno starts with 1, not zero
+                    skip_list = get_rule_skips_from_line(
+                        line=lines[match.lineno - 1],
+                        lintable=file,
+                    )
                     if match.rule.id not in skip_list and match.tag not in skip_list:
                         results.append(match)
         else:
@@ -176,13 +293,15 @@ class JinjaRule(AnsibleLintRule):
 
         self.env.lstrip_blocks = False
         self.env.trim_blocks = False
+        self.env.autoescape = True
+        self.env.newline_sequence = "\n"
         tokens = [
             Token(lineno=t[0], token_type=t[1], value=t[2]) for t in self.env.lex(text)
         ]
         new_text = self.unlex(tokens)
         if text != new_text:
             _logger.debug(
-                "Unable to perform full roundtrip lex-unlex on jinja template (expected when '-' modifier is used): {text} -> {new_text}"
+                "Unable to perform full roundtrip lex-unlex on jinja template (expected when '-' modifier is used): {text} -> {new_text}",
             )
         return tokens
 
@@ -199,9 +318,12 @@ class JinjaRule(AnsibleLintRule):
             last_value = value
         return result
 
-    # pylint: disable=too-many-branches,too-many-statements,too-many-locals
-    def check_whitespace(  # noqa: max-complexity: 13
-        self, text: str, key: str, lintable: Lintable | None = None
+    # pylint: disable=too-many-locals
+    def check_whitespace(
+        self,
+        text: str,
+        key: str,
+        lintable: Lintable | None = None,
     ) -> tuple[str, str, str]:
         """Check spacing inside given jinja2 template string.
 
@@ -211,7 +333,7 @@ class JinjaRule(AnsibleLintRule):
         :returns: (string, string, string)  reformatted text, detailed error, error tag
         """
 
-        def cook(value: str, implicit: bool = False) -> str:
+        def cook(value: str, *, implicit: bool = False) -> str:
             """Prepare an implicit string for jinja parsing when needed."""
             if not implicit:
                 return value
@@ -220,7 +342,7 @@ class JinjaRule(AnsibleLintRule):
                 return value
             return f"{{{{ {value} }}}}"
 
-        def uncook(value: str, implicit: bool = False) -> str:
+        def uncook(value: str, *, implicit: bool = False) -> str:
             """Restore an string to original form when it was an implicit one."""
             if not implicit:
                 return value
@@ -245,13 +367,16 @@ class JinjaRule(AnsibleLintRule):
             implicit = True
             text = cook(text, implicit=implicit)
 
+        # don't try to lex strings that have no jinja inside them
+        if not has_jinja(text):
+            return text, "", "spacing"
+
         expr_str = None
         expr_type = None
         verb_skipped = True
         lineno = 1
         try:
             for token in self.lex(text):
-
                 if (
                     expr_type
                     and expr_type.startswith("{%")
@@ -271,8 +396,9 @@ class JinjaRule(AnsibleLintRule):
                     # process expression
                     # pylint: disable=unsupported-membership-test
                     if isinstance(expr_str, str) and "\n" in expr_str:
-                        raise NotImplementedError()
-                    expr_str = blacken(expr_str)
+                        raise NotImplementedError  # noqa: TRY301
+                    leading_spaces = " " * (len(expr_str) - len(expr_str.lstrip()))
+                    expr_str = leading_spaces + blacken(expr_str.lstrip())
                     if tokens[
                         -1
                     ].token_type != "whitespace" and not expr_str.startswith(" "):
@@ -291,21 +417,19 @@ class JinjaRule(AnsibleLintRule):
 
         except jinja2.exceptions.TemplateSyntaxError as exc:
             return "", str(exc.message), "invalid"
-        # https://github.com/PyCQA/pylint/issues/7433 - py311 only
-        # pylint: disable=c-extension-no-member
-        except (NotImplementedError, black.parsing.InvalidInput) as exc:
+        except (NotImplementedError, ValueError) as exc:
             # black is not able to recognize all valid jinja2 templates, so we
             # just ignore InvalidInput errors.
             # NotImplementedError is raised internally for expressions with
             # newlines, as we decided to not touch them yet.
             # These both are documented as known limitations.
             _logger.debug("Ignored jinja internal error %s", exc)
-            return uncook(text, implicit), "", "spacing"
+            return uncook(text, implicit=implicit), "", "spacing"
 
         # finalize
         reformatted = self.unlex(tokens)
         failed = reformatted != text
-        reformatted = uncook(reformatted, implicit)
+        reformatted = uncook(reformatted, implicit=implicit)
         details = (
             f"Jinja2 template rewrite recommendation: `{reformatted}`."
             if failed
@@ -313,45 +437,104 @@ class JinjaRule(AnsibleLintRule):
         )
         return reformatted, details, "spacing"
 
+    def transform(
+        self,
+        match: MatchError,
+        lintable: Lintable,
+        data: CommentedMap | CommentedSeq | str,
+    ) -> None:
+        """Transform jinja2 errors.
+
+        :param match: MatchError instance
+        :param lintable: Lintable instance
+        :param data: data to transform
+        """
+        if match.tag == "jinja[spacing]":
+            self._transform_spacing(match, data)
+
+    def _transform_spacing(
+        self,
+        match: MatchError,
+        data: CommentedMap | CommentedSeq | str,
+    ) -> None:
+        """Transform jinja2 spacing errors.
+
+        The match error was found on a normalized task so we cannot compare the path
+        instead we only compare the key and value, if the task has 2 identical keys with the
+        exact same jinja spacing issue, we may transform them out of order
+
+        :param match: MatchError instance
+        :param data: data to transform
+        """
+        if not isinstance(match.transform_meta, JinjaRuleTMetaSpacing):
+            return
+        if isinstance(data, str):
+            return
+
+        obj = self.seek(match.yaml_path, data)
+        if obj is None:
+            return
+
+        ignored_keys = ("block", "ansible.builtin.block", "ansible.legacy.block")
+        for key, value, path in nested_items_path(
+            data_collection=obj,
+            ignored_keys=ignored_keys,
+        ):
+            if key == match.transform_meta.key and value == match.transform_meta.value:
+                if not path:
+                    continue
+                for pth in path[:-1]:
+                    try:
+                        obj = obj[pth]
+                    except (KeyError, TypeError) as exc:
+                        err = f"Unable to transform {match.transform_meta}: {exc}"
+                        _logger.error(err)  # noqa: TRY400
+                        return
+                try:
+                    obj[path[-1]][key] = match.transform_meta.fixed
+                    match.fixed = True
+
+                except (KeyError, TypeError) as exc:
+                    err = f"Unable to transform {match.transform_meta}: {exc}"
+                    _logger.error(err)  # noqa: TRY400
+                return
+
 
 def blacken(text: str) -> str:
     """Format Jinja2 template using black."""
     return black.format_str(
-        text, mode=black.FileMode(line_length=sys.maxsize, string_normalization=False)
+        text,
+        mode=black.FileMode(line_length=sys.maxsize, string_normalization=False),
     ).rstrip("\n")
 
 
-if "pytest" in sys.modules:  # noqa: C901
+if "pytest" in sys.modules:
+    from unittest import mock
 
     import pytest
 
-    from ansiblelint.rules import RulesCollection  # pylint: disable=ungrouped-imports
-    from ansiblelint.runner import Runner  # pylint: disable=ungrouped-imports
+    # pylint: disable=ungrouped-imports
+    from ansiblelint.rules import RulesCollection
+    from ansiblelint.runner import Runner
+    from ansiblelint.transformer import Transformer
 
-    @pytest.fixture(name="error_expected_lines")
-    def fixture_error_expected_lines() -> list[int]:
-        """Return list of expected error lines."""
-        return [31, 34, 37, 40, 43, 46, 72]
-
-    # 21 68
-    @pytest.fixture(name="lint_error_lines")
-    def fixture_lint_error_lines() -> list[int]:
-        """Get VarHasSpacesRules linting results on test_playbook."""
-        collection = RulesCollection()
-        collection.register(JinjaRule())
-        lintable = Lintable("examples/playbooks/jinja-spacing.yml")
-        results = Runner(lintable, rules=collection).run()
-        return list(map(lambda item: item.linenumber, results))
-
-    def test_jinja_spacing_playbook(
-        error_expected_lines: list[int], lint_error_lines: list[int]
-    ) -> None:
+    def test_jinja_spacing_playbook() -> None:
         """Ensure that expected error lines are matching found linting error lines."""
         # list unexpected error lines or non-matching error lines
-        error_lines_difference = list(
-            set(error_expected_lines).symmetric_difference(set(lint_error_lines))
-        )
-        assert len(error_lines_difference) == 0
+        lineno_list = [33, 36, 39, 42, 45, 48, 74]
+        lintable = Lintable("examples/playbooks/jinja-spacing.yml")
+        collection = RulesCollection()
+        collection.register(JinjaRule())
+        results = Runner(lintable, rules=collection).run()
+        assert len(results) == len(lineno_list)
+        for index, result in enumerate(results):
+            assert result.tag == "jinja[spacing]"
+            assert result.lineno == lineno_list[index]
+
+        # error_lines_difference = list(
+        #     set(error_expected_lines).symmetric_difference(set(lint_error_lines)),
+        # )
+        # assert len(error_lines_difference) == 0
 
     def test_jinja_spacing_vars() -> None:
         """Ensure that expected error details are matching found linting error details."""
@@ -363,7 +546,7 @@ if "pytest" in sys.modules:  # noqa: C901
         error_expected_lineno = [14, 15, 16, 17, 18, 19, 32]
         assert len(results) == len(error_expected_lineno)
         for idx, err in enumerate(results):
-            assert err.linenumber == error_expected_lineno[idx]
+            assert err.lineno == error_expected_lineno[idx]
 
     @pytest.mark.parametrize(
         ("text", "expected", "tag"),
@@ -377,7 +560,8 @@ if "pytest" in sys.modules:  # noqa: C901
             pytest.param("", "", "spacing", id="1"),
             pytest.param("foo", "foo", "spacing", id="2"),
             pytest.param("{##}", "{# #}", "spacing", id="3"),
-            pytest.param("{#  #}", "{# #}", "spacing", id="4"),
+            # we want to keep leading spaces as they might be needed for complex multiline jinja files
+            pytest.param("{#  #}", "{#  #}", "spacing", id="4"),
             pytest.param(
                 "{{-aaa|xx   }}foo\nbar{#some#}\n{%%}",
                 "{{- aaa | xx }}foo\nbar{# some #}\n{% %}",
@@ -385,7 +569,10 @@ if "pytest" in sys.modules:  # noqa: C901
                 id="5",
             ),
             pytest.param(
-                "Shell with jinja filter", "Shell with jinja filter", "spacing", id="6"
+                "Shell with jinja filter",
+                "Shell with jinja filter",
+                "spacing",
+                id="6",
             ),
             pytest.param(
                 "{{{'dummy_2':1}|true}}",
@@ -409,10 +596,7 @@ if "pytest" in sys.modules:  # noqa: C901
             pytest.param("{{foo(123)}}", "{{ foo(123) }}", "spacing", id="11"),
             pytest.param("{{ foo(a.b.c) }}", "{{ foo(a.b.c) }}", "spacing", id="12"),
             # pytest.param(
-            #     "{{ foo | bool else [ ] }}",
-            #     "{{ foo | bool else [] }}",
             #     "spacing",
-            #     id="13",
             # ),
             pytest.param(
                 "{{foo(x =['server_options'])}}",
@@ -421,7 +605,10 @@ if "pytest" in sys.modules:  # noqa: C901
                 id="14",
             ),
             pytest.param(
-                '{{ [ "host", "NA"] }}', '{{ ["host", "NA"] }}', "spacing", id="15"
+                '{{ [ "host", "NA"] }}',
+                '{{ ["host", "NA"] }}',
+                "spacing",
+                id="15",
             ),
             pytest.param(
                 "{{ {'dummy_2': {'nested_dummy_1': value_1,\n    'nested_dummy_2': value_2}} |\ncombine(dummy_1) }}",
@@ -474,7 +661,10 @@ if "pytest" in sys.modules:  # noqa: C901
             ),
             pytest.param("{{ r(1,[]) }}", "{{ r(1, []) }}", "spacing", id="26"),
             pytest.param(
-                "{{ lookup([ddd ]) }}", "{{ lookup([ddd]) }}", "spacing", id="27"
+                "{{ lookup([ddd ]) }}",
+                "{{ lookup([ddd]) }}",
+                "spacing",
+                id="27",
             ),
             pytest.param(
                 "{{ [ x ] if x is string else x }}",
@@ -483,7 +673,6 @@ if "pytest" in sys.modules:  # noqa: C901
                 id="28",
             ),
             pytest.param(
-                # "{% if a|int <= 8 -%} iptables {%- else -%} iptables-nft {%- endif %}",
                 "{% if a|int <= 8 -%} iptables {%- else -%} iptables-nft {%- endif %}",
                 "{% if a | int <= 8 -%} iptables{%- else -%} iptables-nft{%- endif %}",
                 "spacing",
@@ -533,7 +722,10 @@ if "pytest" in sys.modules:  # noqa: C901
             ),
             pytest.param("{{ a +~'b' }}", "{{ a + ~'b' }}", "spacing", id="36"),
             pytest.param(
-                "{{ (a[: -4] *~ b) }}", "{{ (a[:-4] * ~b) }}", "spacing", id="37"
+                "{{ (a[: -4] *~ b) }}",
+                "{{ (a[:-4] * ~b) }}",
+                "spacing",
+                id="37",
             ),
             pytest.param("{{ [a,~ b] }}", "{{ [a, ~b] }}", "spacing", id="38"),
             # Not supported yet due to being accepted by black:
@@ -558,6 +750,22 @@ if "pytest" in sys.modules:  # noqa: C901
                 "spacing",
                 id="43",
             ),
+            # https://github.com/ansible/ansible-lint/pull/3057
+            # since jinja 3.0.0, \r is converted to \n if the string has jinja in it
+            pytest.param(
+                "{{ 'foo' }}\r{{ 'bar' }}",
+                "{{ 'foo' }}\n{{ 'bar' }}",
+                "spacing",
+                id="44",
+            ),
+            # if we do not have any jinja constructs, we should keep original \r
+            # to match ansible behavior
+            pytest.param(
+                "foo\rbar",
+                "foo\rbar",
+                "spacing",
+                id="45",
+            ),
         ),
     )
     def test_jinja(text: str, expected: str, tag: str) -> None:
@@ -565,7 +773,9 @@ if "pytest" in sys.modules:  # noqa: C901
         rule = JinjaRule()
 
         reformatted, details, returned_tag = rule.check_whitespace(
-            text, key="name", lintable=Lintable("playbook.yml")
+            text,
+            key="name",
+            lintable=Lintable("playbook.yml"),
         )
         assert tag == returned_tag, details
         assert expected == reformatted
@@ -596,7 +806,9 @@ if "pytest" in sys.modules:  # noqa: C901
         # implicit jinja2 are working only inside playbooks and tasks
         lintable = Lintable(name="playbook.yml", kind="playbook")
         reformatted, details, returned_tag = rule.check_whitespace(
-            text, key="when", lintable=lintable
+            text,
+            key="when",
+            lintable=lintable,
         )
         assert tag == returned_tag, details
         assert expected == reformatted
@@ -620,16 +832,72 @@ if "pytest" in sys.modules:  # noqa: C901
         """Tests our ability to spot spacing errors inside jinja2 templates."""
         collection = RulesCollection()
         collection.register(JinjaRule())
-        success = "examples/playbooks/rule-jinja-invalid.yml"
+        success = "examples/playbooks/rule-jinja-fail.yml"
         errs = Runner(success, rules=collection).run()
-        assert len(errs) == 1
-        assert errs[0].tag == "jinja[invalid]"
+        assert len(errs) == 2
+        assert errs[0].tag == "jinja[spacing]"
         assert errs[0].rule.id == "jinja"
+        assert errs[0].lineno == 9
+        assert errs[1].tag == "jinja[invalid]"
+        assert errs[1].rule.id == "jinja"
+        assert errs[1].lineno in [9, 10]  # 2.19 has better line identification
 
     def test_jinja_valid() -> None:
         """Tests our ability to parse jinja, even when variables may not be defined."""
         collection = RulesCollection()
         collection.register(JinjaRule())
-        success = "examples/playbooks/rule-jinja-valid.yml"
+        success = "examples/playbooks/rule-jinja-pass.yml"
         errs = Runner(success, rules=collection).run()
         assert len(errs) == 0
+
+    @mock.patch.dict(os.environ, {"ANSIBLE_LINT_WRITE_TMP": "1"}, clear=True)
+    def test_jinja_transform(
+        config_options: Options,
+        default_rules_collection: RulesCollection,
+    ) -> None:
+        """Test transform functionality for jinja rule."""
+        playbook = Path("examples/playbooks/rule-jinja-before.yml")
+        config_options.write_list = ["all"]
+
+        config_options.lintables = [str(playbook)]
+        runner_result = get_matches(
+            rules=default_rules_collection,
+            options=config_options,
+        )
+        transformer = Transformer(result=runner_result, options=config_options)
+        transformer.run()
+
+        matches = runner_result.matches
+        assert len(matches) == 2
+
+        orig_content = playbook.read_text(encoding="utf-8")
+        expected_content = playbook.with_suffix(
+            f".transformed{playbook.suffix}",
+        ).read_text(encoding="utf-8")
+        transformed_content = playbook.with_suffix(f".tmp{playbook.suffix}").read_text(
+            encoding="utf-8",
+        )
+
+        assert orig_content != transformed_content
+        assert expected_content == transformed_content
+        playbook.with_suffix(f".tmp{playbook.suffix}").unlink()
+
+    def test_jinja_nested_var_errors() -> None:
+        """Tests our ability to handle nested var errors from jinja2 templates."""
+
+        def _do_template(*args, **kwargs):  # type: ignore[no-untyped-def] # Templar.do_template has no type hint
+            data = args[1]
+
+            if data != "{{ 12 | random(seed=inventory_hostname) }}":
+                return do_template(*args, **kwargs)  # type: ignore[no-untyped-call]
+
+            msg = "Unexpected templating type error occurred on (foo): bar"
+            raise AnsibleError(str(msg))  # type: ignore[no-untyped-call]
+
+        do_template = Templar.do_template
+        collection = RulesCollection()
+        collection.register(JinjaRule())
+        lintable = Lintable("examples/playbooks/jinja-nested-vars.yml")
+        with mock.patch.object(Templar, "do_template", _do_template):
+            results = Runner(lintable, rules=collection).run()
+            assert len(results) == 0
